import asyncio
from typing import Union

from v6d3music.processing.yprocess import *

__all__ = ('Job', 'Pool', 'JobDescriptor',)


class Job:
    def __init__(self, future: asyncio.Future[None]) -> None:
        self.future = future

    async def run(self) -> Union['Job', None]:
        raise NotImplementedError


class JobDescriptor:
    async def run(self) -> Union['JobDescriptor', None]:
        raise NotImplementedError

    def wrap(self) -> Job:
        return DescriptorJob(asyncio.Future(), self)

    def at(self, pool: 'Pool') -> 'JDC':
        return JDC(self, pool)


class DescriptorJob(Job):
    def __init__(self, future: asyncio.Future[None], descriptor: JobDescriptor) -> None:
        super().__init__(future)
        self.__descriptor = descriptor

    async def run(self) -> Job | None:
        next_descriptor = await self.__descriptor.run()
        if next_descriptor is None:
            return None
        else:
            return DescriptorJob(self.future, next_descriptor)


class Worker:
    def __init__(self) -> None:
        self.__queue: asyncio.Queue[Job | None] = asyncio.Queue()
        self.__working = False
        self.__busy = 0

    def _put_nowait(self, job: Job | None) -> None:
        self.__queue.put_nowait(job)
        self.__busy += 1

    async def _handle(self, job: Job) -> None:
        try:
            next_job = await job.run()
        except BaseException as e:
            job.future.set_exception(e)
        else:
            if next_job is None:
                job.future.set_result(None)
            else:
                self._put_nowait(next_job)

    async def _safe_handle(self, job: Job) -> None:
        try:
            await self._handle(job)
        except asyncio.InvalidStateError:
            pass

    def _on_shutdown(self) -> bool:
        if self.__queue.empty():
            self.__working = False
            return True
        else:
            self._put_nowait(None)
            return False

    async def _on(self, job: Job | None) -> bool:
        if job is None:
            return self._on_shutdown()
        else:
            await self._handle(job)
            return False

    async def _step(self) -> bool:
        job = await self.__queue.get()
        try:
            return await self._on(job)
        finally:
            self.__busy -= 1
            self.__queue.task_done()

    async def _work(self) -> None:
        while True:
            if await self._step():
                return

    def _cancel_job(self, job: Job) -> None:
        try:
            job.future.set_exception(RuntimeError('task left in the worker after shutdown'))
        except asyncio.InvalidStateError:
            pass

    def _cancel_one(self, job: Job | None) -> None:
        if job is not None:
            self._cancel_job(job)

    def _cancel_get(self) -> None:
        job = self.__queue.get_nowait()
        try:
            self._cancel_one(job)
        finally:
            self.__queue.task_done()

    def _cancel_all(self) -> None:
        while not self.__queue.empty():
            self._cancel_get()

    async def _task(self) -> None:
        await self._work()
        if self.__working:
            raise RuntimeError('worker left seemingly running after shutdown')
        if not self.__queue.empty():
            raise RuntimeError('worker failed to finish all its jobs')

    def start(self) -> asyncio.Future[None]:
        if self.__working:
            raise RuntimeError('starting an already running worker')
        self.__working = True
        return asyncio.create_task(self._task())

    def submit(self, job: Job | None) -> None:
        if not self.__working:
            raise RuntimeError('submitting to a non-working worker')
        self._put_nowait(job)

    def working(self) -> bool:
        return self.__working

    def busy(self) -> int:
        return self.__busy


class Working:
    def __init__(self, worker: Worker, task: asyncio.Future[None]) -> None:
        self.__worker = worker
        self.__task = task

    async def close(self) -> None:
        self.__worker.submit(None)
        await self.__task

    def submit(self, job: Job) -> None:
        self.__worker.submit(job)

    @classmethod
    def start(cls) -> 'Working':
        worker = Worker()
        task = worker.start()
        return cls(worker, task)

    def working(self) -> bool:
        return self.__worker.working()

    def busy(self) -> int:
        return self.__worker.busy()


class Pool:
    def __init__(self, workers: int) -> None:
        if workers < 1:
            raise ValueError('non-positive number of workers')
        self.__workers = workers
        self.__working = False
        self.__open = False

    async def __aenter__(self) -> 'Pool':
        if self.__open:
            raise RuntimeError('starting an already open pool')
        if self.__working:
            raise RuntimeError('starting an already running pool')
        self.__working = True
        self.__pool = set(Working.start() for _ in range(self.__workers))
        self.__open = True
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if not self.__working:
            raise RuntimeError('stopping a non-working pool')
        if not self.__open:
            raise RuntimeError('stopping a closed pool')
        self.__open = False
        for working in self.__pool:
            await working.close()
        del self.__pool
        self.__working = False

    def submit(self, job: Job) -> None:
        if not self.__working:
            raise RuntimeError('submitting to a non-working pool')
        if not self.__open:
            raise RuntimeError('submitting to a closed pool')
        min(self.__pool, key=lambda working: working.busy()).submit(job)

    def workers(self) -> int:
        return self.__workers


class JDC:
    def __init__(self, descriptor: JobDescriptor, pool: Pool) -> None:
        self.__descriptor = descriptor
        self.__pool = pool

    async def __aenter__(self) -> 'JDC':
        job = self.__descriptor.wrap()
        self.__future = job.future
        self.__pool.submit(job)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.__future
